@@ -18,8 +18,6 @@
  * Or go to http://www.gnu.org/copyleft/lgpl.html
  */
 
-#define _CRT_SECURE_NO_DEPRECATE // get rid of sprintf security warnings on VS2005
-
 #include "config.h"
 
 #include <math.h>
@@ -33,16 +31,11 @@
 #include "AL/alc.h"
 #include "alThunk.h"
 #include "alSource.h"
+#include "alBuffer.h"
 #include "alExtension.h"
 #include "alAuxEffectSlot.h"
 #include "bs2b.h"
-
-///////////////////////////////////////////////////////
-// DEBUG INFORMATION
-
-char _alDebug[256];
-
-///////////////////////////////////////////////////////
+#include "alu.h"
 
 
 #define EmptyFuncs { NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL }
@@ -57,12 +50,21 @@ static struct {
 #ifdef HAVE_OSS
     { "oss", alc_oss_init, EmptyFuncs },
 #endif
+#ifdef HAVE_SOLARIS
+    { "solaris", alc_solaris_init, EmptyFuncs },
+#endif
 #ifdef HAVE_DSOUND
     { "dsound", alcDSoundInit, EmptyFuncs },
 #endif
 #ifdef HAVE_WINMM
     { "winmm", alcWinMMInit, EmptyFuncs },
 #endif
+#ifdef HAVE_PORTAUDIO
+    { "port", alc_pa_init, EmptyFuncs },
+#endif
+#ifdef HAVE_PULSEAUDIO
+    { "pulse", alc_pulse_init, EmptyFuncs },
+#endif
 
     { "wave", alc_wave_init, EmptyFuncs },
 
@@ -182,6 +184,8 @@ static ALCint alcEFXMinorVersion = 0;
 static ALCdevice *g_pDeviceList = NULL;
 static ALCuint    g_ulDeviceCount = 0;
 
+static CRITICAL_SECTION g_csMutex;
+
 // Context List
 static ALCcontext *g_pContextList = NULL;
 static ALCuint     g_ulContextCount = 0;
@@ -189,23 +193,69 @@ static ALCuint     g_ulContextCount = 0;
 // Context Error
 static ALCenum g_eLastContextError = ALC_NO_ERROR;
 
+static ALboolean init_done = AL_FALSE;
+
 ///////////////////////////////////////////////////////
 
 
 ///////////////////////////////////////////////////////
 // ALC Related helper functions
+#ifdef _WIN32
+BOOL APIENTRY DllMain(HANDLE hModule,DWORD ul_reason_for_call,LPVOID lpReserved)
+{
+    (void)lpReserved;
+
+    // Perform actions based on the reason for calling.
+    switch(ul_reason_for_call)
+    {
+        case DLL_PROCESS_ATTACH:
+            DisableThreadLibraryCalls(hModule);
+            break;
+
+        case DLL_PROCESS_DETACH:
+            if(!init_done)
+                break;
+            ReleaseALC();
+            ReleaseALBuffers();
+            ReleaseALEffects();
+            ReleaseALFilters();
+            FreeALConfig();
+            ALTHUNK_EXIT();
+            DeleteCriticalSection(&g_csMutex);
+            break;
+    }
+    return TRUE;
+}
+#else
+#ifdef HAVE_GCC_DESTRUCTOR
+static void my_deinit() __attribute__((destructor));
+static void my_deinit()
+{
+    static ALenum once = AL_FALSE;
+    if(once || !init_done) return;
+    once = AL_TRUE;
+
+    ReleaseALC();
+    ReleaseALBuffers();
+    ReleaseALEffects();
+    ReleaseALFilters();
+    FreeALConfig();
+    ALTHUNK_EXIT();
+    DeleteCriticalSection(&g_csMutex);
+}
+#endif
+#endif
 
 static void InitAL(void)
 {
-    static int done = 0;
-    if(!done)
+    if(!init_done)
     {
         int i;
         const char *devs, *str;
 
-        done = 1;
+        init_done = AL_TRUE;
 
-        InitializeCriticalSection(&_alMutex);
+        InitializeCriticalSection(&g_csMutex);
         ALTHUNK_INIT();
         ReadALConfig();
 
@@ -257,6 +307,39 @@ static void InitAL(void)
                            strcasecmp(str, "yes") == 0 ||
                            strcasecmp(str, "on") == 0 ||
                            atoi(str) != 0);
+
+        str = GetConfigValue(NULL, "excludefx", "");
+        if(str[0])
+        {
+            const struct {
+                const char *name;
+                int type;
+            } EffectList[] = {
+                { "eaxreverb", EAXREVERB },
+                { "reverb", REVERB },
+                { "echo", ECHO },
+                { NULL, 0 }
+            };
+            int n;
+            size_t len;
+            const char *next = str;
+
+            do {
+                str = next;
+                next = strchr(str, ',');
+
+                if(!str[0] || next == str)
+                    continue;
+
+                len = (next ? ((size_t)(next-str)) : strlen(str));
+                for(n = 0;EffectList[n].name;n++)
+                {
+                    if(len == strlen(EffectList[n].name) &&
+                       strncmp(EffectList[n].name, str, len) == 0)
+                        DisabledEffects[EffectList[n].type] = AL_TRUE;
+                }
+            } while(next++);
+        }
     }
 }
 
@@ -335,7 +418,7 @@ ALCvoid SetALCError(ALenum errorCode)
 ALCvoid SuspendContext(ALCcontext *pContext)
 {
     (void)pContext;
-    EnterCriticalSection(&_alMutex);
+    EnterCriticalSection(&g_csMutex);
 }
 
 
@@ -347,7 +430,7 @@ ALCvoid SuspendContext(ALCcontext *pContext)
 ALCvoid ProcessContext(ALCcontext *pContext)
 {
     (void)pContext;
-    LeaveCriticalSection(&_alMutex);
+    LeaveCriticalSection(&g_csMutex);
 }
 
 
@@ -392,7 +475,11 @@ static ALvoid InitContext(ALCcontext *pContext)
     pContext->lNumStereoSources = 1;
     pContext->lNumMonoSources = pContext->Device->MaxNoOfSources - pContext->lNumStereoSources;
 
-    strcpy(pContext->ExtensionList, "AL_EXT_EXPONENT_DISTANCE AL_EXT_FLOAT32 AL_EXT_IMA4 AL_EXT_LINEAR_DISTANCE AL_EXT_MCFORMATS AL_EXT_OFFSET AL_LOKI_quadriphonic");
+    pContext->NumSends = GetConfigValueInt(NULL, "sends", MAX_SENDS);
+    if(pContext->NumSends > MAX_SENDS)
+        pContext->NumSends = MAX_SENDS;
+
+    pContext->ExtensionList = "AL_EXTX_buffer_sub_data AL_EXT_EXPONENT_DISTANCE AL_EXT_FLOAT32 AL_EXT_IMA4 AL_EXT_LINEAR_DISTANCE AL_EXT_MCFORMATS AL_EXT_OFFSET AL_EXTX_source_distance_model AL_LOKI_quadriphonic";
 
     level = GetConfigValueInt(NULL, "cf_level", 0);
     if(level > 0 && level <= 6)
@@ -401,6 +488,8 @@ static ALvoid InitContext(ALCcontext *pContext)
         bs2b_set_srate(pContext->bs2b, pContext->Frequency);
         bs2b_set_level(pContext->bs2b, level);
     }
+
+    aluInitPanning(pContext);
 }
 
 
@@ -435,44 +524,46 @@ ALCAPI ALCdevice* ALCAPIENTRY alcCaptureOpenDevice(const ALCchar *deviceName, AL
 
     InitAL();
 
+    if(SampleSize <= 0)
+    {
+        SetALCError(ALC_INVALID_VALUE);
+        return NULL;
+    }
+
     if(deviceName && !deviceName[0])
         deviceName = NULL;
 
     pDevice = malloc(sizeof(ALCdevice));
     if (pDevice)
     {
-        if (SampleSize > 0)
+        //Initialise device structure
+        memset(pDevice, 0, sizeof(ALCdevice));
+
+        //Validate device
+        pDevice->IsCaptureDevice = AL_TRUE;
+
+        pDevice->Frequency = frequency;
+        pDevice->Format = format;
+
+        for(i = 0;BackendList[i].Init;i++)
         {
-            //Initialise device structure
-            memset(pDevice, 0, sizeof(ALCdevice));
-
-            //Validate device
-            pDevice->IsCaptureDevice = AL_TRUE;
-
-            pDevice->Frequency = frequency;
-            pDevice->Format = format;
-
-            for(i = 0;BackendList[i].Init;i++)
+            pDevice->Funcs = &BackendList[i].Funcs;
+            if(ALCdevice_OpenCapture(pDevice, deviceName, frequency, format, SampleSize))
             {
-                pDevice->Funcs = &BackendList[i].Funcs;
-                if(ALCdevice_OpenCapture(pDevice, deviceName, frequency, format, SampleSize))
-                {
-                    SuspendContext(NULL);
-                    pDevice->next = g_pDeviceList;
-                    g_pDeviceList = pDevice;
-                    g_ulDeviceCount++;
-                    ProcessContext(NULL);
+                SuspendContext(NULL);
+                pDevice->next = g_pDeviceList;
+                g_pDeviceList = pDevice;
+                g_ulDeviceCount++;
+                ProcessContext(NULL);
 
-                    DeviceFound = ALC_TRUE;
-                    break;
-                }
+                DeviceFound = ALC_TRUE;
+                break;
             }
         }
-        else
-            SetALCError(ALC_INVALID_VALUE);
 
         if(!DeviceFound)
         {
+            SetALCError(ALC_INVALID_VALUE);
             free(pDevice);
             pDevice = NULL;
         }
@@ -725,6 +816,8 @@ ALCAPI ALCvoid ALCAPIENTRY alcGetIntegerv(ALCdevice *device,ALCenum param,ALsize
                 case ALC_MAX_AUXILIARY_SENDS:
                     if(!size)
                         SetALCError(ALC_INVALID_VALUE);
+                    else if(device && device->Context)
+                        *data = device->Context->NumSends;
                     else
                         *data = MAX_SENDS;
                     break;
@@ -766,7 +859,7 @@ ALCAPI ALCvoid ALCAPIENTRY alcGetIntegerv(ALCdevice *device,ALCenum param,ALsize
                             data[i++] = device->Context->lNumStereoSources;
 
                             data[i++] = ALC_MAX_AUXILIARY_SENDS;
-                            data[i++] = MAX_SENDS;
+                            data[i++] = device->Context->NumSends;
                         }
                         ProcessContext(NULL);
 
@@ -931,6 +1024,7 @@ ALCAPI ALCcontext* ALCAPIENTRY alcCreateContext(ALCdevice *device, const ALCint
 {
     ALCcontext *ALContext = NULL;
     ALuint      ulAttributeIndex, ulRequestedStereoSources;
+    ALuint      RequestedSends;
 
     if ((device)&&(!device->IsCaptureDevice))
     {
@@ -960,7 +1054,7 @@ ALCAPI ALCcontext* ALCAPIENTRY alcCreateContext(ALCdevice *device, const ALCint
 
             ProcessContext(NULL);
 
-            // Check for Voice Count attributes
+            // Check for attributes
             if (attrList)
             {
                 ulAttributeIndex = 0;
@@ -975,7 +1069,16 @@ ALCAPI ALCcontext* ALCAPIENTRY alcCreateContext(ALCdevice *device, const ALCint
 
                         ALContext->lNumStereoSources = ulRequestedStereoSources;
                         ALContext->lNumMonoSources = ALContext->Device->MaxNoOfSources - ALContext->lNumStereoSources;
-                        break;
+                    }
+
+                    if(attrList[ulAttributeIndex] == ALC_MAX_AUXILIARY_SENDS)
+                    {
+                        RequestedSends = attrList[ulAttributeIndex + 1];
+
+                        if(RequestedSends > ALContext->NumSends)
+                            RequestedSends = ALContext->NumSends;
+
+                        ALContext->NumSends = RequestedSends;
                     }
 
                     ulAttributeIndex += 2;
@@ -1169,38 +1272,45 @@ ALCAPI ALCdevice* ALCAPIENTRY alcOpenDevice(const ALCchar *deviceName)
         if(!aluChannelsFromFormat(device->Format))
             device->Format = AL_FORMAT_STEREO16;
 
-        device->UpdateSize = GetConfigValueInt(NULL, "refresh", 8192);
+        device->UpdateSize = GetConfigValueInt(NULL, "refresh", 4096);
         if((ALint)device->UpdateSize <= 0)
-            device->UpdateSize = 8192;
+            device->UpdateSize = 4096;
 
         device->MaxNoOfSources = GetConfigValueInt(NULL, "sources", 256);
         if((ALint)device->MaxNoOfSources <= 0)
             device->MaxNoOfSources = 256;
 
+        device->AuxiliaryEffectSlotMax = GetConfigValueInt(NULL, "slots", 4);
+        if((ALint)device->AuxiliaryEffectSlotMax <= 0)
+            device->AuxiliaryEffectSlotMax = 4;
+
         // Find a playback device to open
+        SuspendContext(NULL);
         for(i = 0;BackendList[i].Init;i++)
         {
             device->Funcs = &BackendList[i].Funcs;
             if(ALCdevice_OpenPlayback(device, deviceName))
             {
-                SuspendContext(NULL);
                 device->next = g_pDeviceList;
                 g_pDeviceList = device;
                 g_ulDeviceCount++;
-                ProcessContext(NULL);
 
                 bDeviceFound = AL_TRUE;
                 break;
             }
         }
+        ProcessContext(NULL);
 
         if (!bDeviceFound)
         {
             // No suitable output device found
+            SetALCError(ALC_INVALID_VALUE);
             free(device);
             device = NULL;
         }
     }
+    else
+        SetALCError(ALC_OUT_OF_MEMORY);
 
     return device;
 }
@@ -1230,7 +1340,12 @@ ALCAPI ALCboolean ALCAPIENTRY alcCloseDevice(ALCdevice *pDevice)
         ProcessContext(NULL);
 
         if(pDevice->Context)
+        {
+#ifdef _DEBUG
+            AL_PRINT("alcCloseDevice(): destroying 1 Context\n");
+#endif
             alcDestroyContext(pDevice->Context);
+        }
         ALCdevice_ClosePlayback(pDevice);
 
         //Release device structure
@@ -1248,11 +1363,9 @@ ALCAPI ALCboolean ALCAPIENTRY alcCloseDevice(ALCdevice *pDevice)
 
 ALCvoid ReleaseALC(ALCvoid)
 {
-    InitAL();
-
 #ifdef _DEBUG
-    if(g_ulContextCount > 0)
-        AL_PRINT("exit() %u device(s) and %u context(s) NOT deleted\n", g_ulDeviceCount, g_ulContextCount);
+    if(g_ulDeviceCount > 0)
+        AL_PRINT("exit(): closing %u Device%s\n", g_ulDeviceCount, (g_ulDeviceCount>1)?"s":"");
 #endif
 
     while(g_pDeviceList)
