@@ -18,11 +18,14 @@ typedef struct ALbuffer_struct
     ALshort *data;
     ALsizei  size;
     ALsizei  frequency;
+    ALsizei  padding;
     ALenum   state;
     ALuint   refcount; // Number of sources using this buffer (deletion can only occur when this is 0)
     struct ALbuffer_struct *next;
 } ALbuffer;
 
+ALvoid ALAPIENTRY alBufferSubDataEXT(ALuint buffer,ALenum format,const ALvoid *data,ALsizei offset,ALsizei length);
+
 ALvoid ReleaseALBuffers(ALvoid);
 
 #ifdef __cplusplus
