@@ -18,8 +18,6 @@
  * Or go to http://www.gnu.org/copyleft/lgpl.html
  */
 
-#define _CRT_SECURE_NO_DEPRECATE // get rid of sprintf security warnings on VS2005
-
 #include "config.h"
 
 #include <stdlib.h>
@@ -34,6 +32,9 @@
 
 
 static void LoadData(ALbuffer *ALBuf, const ALubyte *data, ALsizei size, ALuint freq, ALenum OrigFormat, ALenum NewFormat);
+static void ConvertData(ALshort *dst, const ALvoid *src, ALint origBytes, ALsizei len);
+static void ConvertDataRear(ALshort *dst, const ALvoid *src, ALint origBytes, ALsizei len);
+static void ConvertDataIMA4(ALshort *dst, const ALvoid *src, ALint origChans, ALsizei len);
 
 /*
  *  AL Buffer Functions
@@ -142,63 +143,59 @@ ALAPI ALvoid ALAPIENTRY alDeleteBuffers(ALsizei n, const ALuint *puiBuffers)
     // Check we are actually Deleting some Buffers
     if (n >= 0)
     {
-        if ((ALuint)n <= g_uiBufferCount)
+        // Check that all the buffers are valid and can actually be deleted
+        for (i = 0; i < n; i++)
         {
-            // Check that all the buffers are valid and can actually be deleted
-            for (i = 0; i < n; i++)
+            // Check for valid Buffer ID (can be NULL buffer)
+            if (alIsBuffer(puiBuffers[i]))
             {
-                // Check for valid Buffer ID (can be NULL buffer)
-                if (alIsBuffer(puiBuffers[i]))
+                // If not the NULL buffer, check that the reference count is 0
+                ALBuf = ((ALbuffer *)ALTHUNK_LOOKUPENTRY(puiBuffers[i]));
+                if (ALBuf)
                 {
-                    // If not the NULL buffer, check that the reference count is 0
-                    ALBuf = ((ALbuffer *)ALTHUNK_LOOKUPENTRY(puiBuffers[i]));
-                    if (ALBuf)
+                    if (ALBuf->refcount != 0)
                     {
-                        if (ALBuf->refcount != 0)
-                        {
-                            // Buffer still in use, cannot be deleted
-                            alSetError(AL_INVALID_OPERATION);
-                            bFailed = AL_TRUE;
-                        }
+                        // Buffer still in use, cannot be deleted
+                        alSetError(AL_INVALID_OPERATION);
+                        bFailed = AL_TRUE;
                     }
                 }
-                else
-                {
-                    // Invalid Buffer
-                    alSetError(AL_INVALID_NAME);
-                    bFailed = AL_TRUE;
-                }
             }
-
-            // If all the Buffers were valid (and have Reference Counts of 0), then we can delete them
-            if (!bFailed)
+            else
             {
-                for (i = 0; i < n; i++)
+                // Invalid Buffer
+                alSetError(AL_INVALID_NAME);
+                bFailed = AL_TRUE;
+            }
+        }
+
+        // If all the Buffers were valid (and have Reference Counts of 0), then we can delete them
+        if (!bFailed)
+        {
+            for (i = 0; i < n; i++)
+            {
+                if (puiBuffers[i] && alIsBuffer(puiBuffers[i]))
                 {
+                    ALbuffer **list = &g_pBuffers;
+
                     ALBuf=((ALbuffer *)ALTHUNK_LOOKUPENTRY(puiBuffers[i]));
-                    if (ALBuf)
-                    {
-                        ALbuffer **list = &g_pBuffers;
-                        while(*list && *list != ALBuf)
-                            list = &(*list)->next;
+                    while(*list && *list != ALBuf)
+                        list = &(*list)->next;
 
-                        if(*list)
-                            *list = (*list)->next;
+                    if(*list)
+                        *list = (*list)->next;
 
-                        // Release the memory used to store audio data
-                        free(ALBuf->data);
+                    // Release the memory used to store audio data
+                    free(ALBuf->data);
 
-                        // Release buffer structure
-                        ALTHUNK_REMOVEENTRY(puiBuffers[i]);
-                        memset(ALBuf, 0, sizeof(ALbuffer));
-                        g_uiBufferCount--;
-                        free(ALBuf);
-                    }
+                    // Release buffer structure
+                    ALTHUNK_REMOVEENTRY(puiBuffers[i]);
+                    memset(ALBuf, 0, sizeof(ALbuffer));
+                    g_uiBufferCount--;
+                    free(ALBuf);
                 }
             }
         }
-        else
-            alSetError(AL_INVALID_NAME);
     }
     else
         alSetError(AL_INVALID_VALUE);
@@ -259,14 +256,10 @@ ALAPI ALboolean ALAPIENTRY alIsBuffer(ALuint uiBuffer)
 */
 ALAPI ALvoid ALAPIENTRY alBufferData(ALuint buffer,ALenum format,const ALvoid *data,ALsizei size,ALsizei freq)
 {
-    ALuint *IMAData,IMACode;
     ALCcontext *Context;
-    ALint Sample,Index;
-    ALint LeftSample,LeftIndex;
-    ALint RightSample,RightIndex;
-    ALuint LeftIMACode,RightIMACode;
+    ALsizei padding = 2;
     ALbuffer *ALBuf;
-    ALsizei i,j,k;
+    ALvoid *temp;
 
     Context = alcGetCurrentContext();
     SuspendContext(Context);
@@ -298,107 +291,35 @@ ALAPI ALvoid ALAPIENTRY alBufferData(ALuint buffer,ALenum format,const ALvoid *d
                     ALuint OrigBytes = ((format==AL_FORMAT_REAR8) ? 1 :
                                         ((format==AL_FORMAT_REAR16) ? 2 :
                                          4));
-                    ALsizei i;
 
                     assert(aluBytesFromFormat(NewFormat) == 2);
 
-                    if ((size%(OrigBytes*2)) != 0)
+                    if((size%(OrigBytes*2)) != 0)
                     {
                         alSetError(AL_INVALID_VALUE);
                         break;
                     }
 
-                    switch(OrigBytes)
+                    size /= OrigBytes;
+                    size *= 2;
+
+                    // Samples are converted to 16 bit here
+                    temp = realloc(ALBuf->data, (padding*NewChannels + size) * sizeof(ALshort));
+                    if(temp)
                     {
-                    case 1:
-                        size /= sizeof(ALubyte);
-                        size *= 2;
+                        ALBuf->data = temp;
+                        ConvertDataRear(ALBuf->data, data, OrigBytes, size);
 
-                        // 8bit Samples are converted to 16 bit here
-                        // Allocate 8 extra samples
-                        ALBuf->data = realloc(ALBuf->data, (8*NewChannels + size) * (1*sizeof(ALshort)));
-                        if (ALBuf->data)
-                        {
-                            for (i = 0;i < size;i+=4)
-                            {
-                                ALBuf->data[i+0] = 0;
-                                ALBuf->data[i+1] = 0;
-                                ALBuf->data[i+2] = (ALshort)((((ALubyte*)data)[i/2+0]-128) << 8);
-                                ALBuf->data[i+3] = (ALshort)((((ALubyte*)data)[i/2+1]-128) << 8);
-                            }
-                            memset(&(ALBuf->data[size]), 0, 16*NewChannels);
+                        memset(&(ALBuf->data[size]), 0, padding*NewChannels*sizeof(ALshort));
 
-                            ALBuf->format = NewFormat;
-                            ALBuf->eOriginalFormat = format;
-                            ALBuf->size = size*1*sizeof(ALshort);
-                            ALBuf->frequency = freq;
-                        }
-                        else
-                            alSetError(AL_OUT_OF_MEMORY);
-                        break;
-
-                    case 2:
-                        size /= sizeof(ALshort);
-                        size *= 2;
-
-                        // Allocate 8 extra samples
-                        ALBuf->data = realloc(ALBuf->data, (8*NewChannels + size) * (1*sizeof(ALshort)));
-                        if (ALBuf->data)
-                        {
-                            for (i = 0;i < size;i+=4)
-                            {
-                                ALBuf->data[i+0] = 0;
-                                ALBuf->data[i+1] = 0;
-                                ALBuf->data[i+2] = ((ALshort*)data)[i/2+0];
-                                ALBuf->data[i+3] = ((ALshort*)data)[i/2+1];
-                            }
-                            memset(&(ALBuf->data[size]), 0, 16*NewChannels);
-
-                            ALBuf->format = NewFormat;
-                            ALBuf->eOriginalFormat = format;
-                            ALBuf->size = size*1*sizeof(ALshort);
-                            ALBuf->frequency = freq;
-                        }
-                        else
-                            alSetError(AL_OUT_OF_MEMORY);
-                        break;
-
-                    case 4:
-                        size /= sizeof(ALfloat);
-                        size *= 2;
-
-                        // Allocate 8 extra samples
-                        ALBuf->data = realloc(ALBuf->data, (8*NewChannels + size) * (1*sizeof(ALshort)));
-                        if (ALBuf->data)
-                        {
-                            ALint smp;
-                            for (i = 0;i < size;i+=4)
-                            {
-                                ALBuf->data[i+0] = 0;
-                                ALBuf->data[i+1] = 0;
-                                smp = (((ALfloat*)data)[i/2+0] * 32767.5f - 0.5);
-                                smp = min(smp,  32767);
-                                smp = max(smp, -32768);
-                                ALBuf->data[i+2] = (ALshort)smp;
-                                smp = (((ALfloat*)data)[i/2+1] * 32767.5f - 0.5);
-                                smp = min(smp,  32767);
-                                smp = max(smp, -32768);
-                                ALBuf->data[i+3] = (ALshort)smp;
-                            }
-                            memset(&(ALBuf->data[size]), 0, 16*NewChannels);
-
-                            ALBuf->format = NewFormat;
-                            ALBuf->eOriginalFormat = format;
-                            ALBuf->size = size*1*sizeof(ALshort);
-                            ALBuf->frequency = freq;
-                        }
-                        else
-                            alSetError(AL_OUT_OF_MEMORY);
-                        break;
-
-                    default:
-                        assert(0);
+                        ALBuf->format = NewFormat;
+                        ALBuf->eOriginalFormat = format;
+                        ALBuf->size = size*sizeof(ALshort);
+                        ALBuf->frequency = freq;
+                        ALBuf->padding = padding;
                     }
+                    else
+                        alSetError(AL_OUT_OF_MEMORY);
                 }   break;
 
                 case AL_FORMAT_QUAD8_LOKI:
@@ -428,157 +349,39 @@ ALAPI ALvoid ALAPIENTRY alBufferData(ALuint buffer,ALenum format,const ALvoid *d
                     break;
 
                 case AL_FORMAT_MONO_IMA4:
+                case AL_FORMAT_STEREO_IMA4: {
+                    int OrigChans = ((format==AL_FORMAT_MONO_IMA4) ? 1 : 2);
+
                     // Here is where things vary:
-                    // nVidia and Apple use 64+1 samples per block => block_size=36 bytes
-                    // Most PC sound software uses 2040+1 samples per block -> block_size=1024 bytes
-                    if ((size%36) == 0)
+                    // nVidia and Apple use 64+1 samples per channel per block => block_size=36*chans bytes
+                    // Most PC sound software uses 2040+1 samples per channel per block -> block_size=1024*chans bytes
+                    if((size%(36*OrigChans)) != 0)
                     {
-                        // Allocate 8 extra samples (16 bytes)
-                        ALBuf->data=realloc(ALBuf->data,16+(size/36)*(65*sizeof(ALshort)));
-                        if (ALBuf->data)
-                        {
-                            ALBuf->format = AL_FORMAT_MONO16;
-                            ALBuf->eOriginalFormat = AL_FORMAT_MONO_IMA4;
-                            IMAData=(ALuint *)data;
-                            for (i=0;i<size/36;i++)
-                            {
-                                Sample=((ALshort *)IMAData)[0];
-                                Index=((ALshort *)IMAData)[1];
-
-                                Index=Index<0?0:Index;
-                                Index=Index>88?88:Index;
-
-                                ALBuf->data[i*65]=(short)Sample;
-
-                                IMAData++;
-
-                                for (j=1;j<65;j+=8)
-                                {
-                                    IMACode=*IMAData;
-                                    for (k=0;k<8;k+=2)
-                                    {
-                                        Sample+=((g_IMAStep_size[Index]*g_IMACodeword_4[IMACode&15])/8);
-                                        Index+=g_IMAIndex_adjust_4[IMACode&15];
-                                        if (Sample<-32768) Sample=-32768;
-                                        else if (Sample>32767) Sample=32767;
-                                        if (Index<0) Index=0;
-                                        else if (Index>88) Index=88;
-                                        ALBuf->data[i*65+j+k]=(short)Sample;
-                                        IMACode>>=4;
-
-                                        Sample+=((g_IMAStep_size[Index]*g_IMACodeword_4[IMACode&15])/8);
-                                        Index+=g_IMAIndex_adjust_4[IMACode&15];
-                                        if (Sample<-32768) Sample=-32768;
-                                        else if (Sample>32767) Sample=32767;
-                                        if (Index<0) Index=0;
-                                        else if (Index>88) Index=88;
-                                        ALBuf->data[i*65+j+k+1]=(short)Sample;
-                                        IMACode>>=4;
-                                    }
-                                    IMAData++;
-                                }
-                            }
-                            memset(&(ALBuf->data[(size/36*65)]), 0, 16);
-                            ALBuf->size=size/36*65*sizeof(ALshort);
-                            ALBuf->frequency=freq;
-                        }
-                        else
-                            alSetError(AL_OUT_OF_MEMORY);
-                    }
-                    else
-                         alSetError(AL_INVALID_VALUE);
-                    break;
-
-                case AL_FORMAT_STEREO_IMA4:
-                    // Here is where things vary:
-                    // nVidia and Apple use 64+1 samples per channel per block => block_size=72 bytes
-                    // Most PC sound software uses 2040+1 samples per channel per block -> block_size=2048 bytes
-                    if ((size%72) == 0)
-                    {
-                        // Allocate 8 extra samples (32 bytes)
-                        ALBuf->data=realloc(ALBuf->data,32+(size/72)*(2*65*sizeof(ALshort)));
-                        if (ALBuf->data)
-                        {
-                            ALBuf->format = AL_FORMAT_STEREO16;
-                            ALBuf->eOriginalFormat = AL_FORMAT_STEREO_IMA4;
-                            IMAData=(ALuint *)data;
-                            for (i=0;i<size/72;i++)
-                            {
-                                LeftSample=((ALshort *)IMAData)[0];
-                                LeftIndex=((ALshort *)IMAData)[1];
-
-                                LeftIndex=LeftIndex<0?0:LeftIndex;
-                                LeftIndex=LeftIndex>88?88:LeftIndex;
-
-                                ALBuf->data[i*2*65]=(short)LeftSample;
-
-                                IMAData++;
-
-                                RightSample=((ALshort *)IMAData)[0];
-                                RightIndex=((ALshort *)IMAData)[1];
-
-                                RightIndex=RightIndex<0?0:RightIndex;
-                                RightIndex=RightIndex>88?88:RightIndex;
-
-                                ALBuf->data[i*2*65+1]=(short)RightSample;
-
-                                IMAData++;
-
-                                for (j=2;j<130;j+=16)
-                                {
-                                    LeftIMACode=IMAData[0];
-                                    RightIMACode=IMAData[1];
-                                    for (k=0;k<16;k+=4)
-                                    {
-                                        LeftSample+=((g_IMAStep_size[LeftIndex]*g_IMACodeword_4[LeftIMACode&15])/8);
-                                        LeftIndex+=g_IMAIndex_adjust_4[LeftIMACode&15];
-                                        if (LeftSample<-32768) LeftSample=-32768;
-                                        else if (LeftSample>32767) LeftSample=32767;
-                                        if (LeftIndex<0) LeftIndex=0;
-                                        else if (LeftIndex>88) LeftIndex=88;
-                                        ALBuf->data[i*2*65+j+k]=(short)LeftSample;
-                                        LeftIMACode>>=4;
-
-                                        RightSample+=((g_IMAStep_size[RightIndex]*g_IMACodeword_4[RightIMACode&15])/8);
-                                        RightIndex+=g_IMAIndex_adjust_4[RightIMACode&15];
-                                        if (RightSample<-32768) RightSample=-32768;
-                                        else if (RightSample>32767) RightSample=32767;
-                                        if (RightIndex<0) RightIndex=0;
-                                        else if (RightIndex>88) RightIndex=88;
-                                        ALBuf->data[i*2*65+j+k+1]=(short)RightSample;
-                                        RightIMACode>>=4;
-
-                                        LeftSample+=((g_IMAStep_size[LeftIndex]*g_IMACodeword_4[LeftIMACode&15])/8);
-                                        LeftIndex+=g_IMAIndex_adjust_4[LeftIMACode&15];
-                                        if (LeftSample<-32768) LeftSample=-32768;
-                                        else if (LeftSample>32767) LeftSample=32767;
-                                        if (LeftIndex<0) LeftIndex=0;
-                                        else if (LeftIndex>88) LeftIndex=88;
-                                        ALBuf->data[i*2*65+j+k+2]=(short)LeftSample;
-                                        LeftIMACode>>=4;
-
-                                        RightSample+=((g_IMAStep_size[RightIndex]*g_IMACodeword_4[RightIMACode&15])/8);
-                                        RightIndex+=g_IMAIndex_adjust_4[RightIMACode&15];
-                                        if (RightSample<-32768) RightSample=-32768;
-                                        else if (RightSample>32767) RightSample=32767;
-                                        if (RightIndex<0) RightIndex=0;
-                                        else if (RightIndex>88) RightIndex=88;
-                                        ALBuf->data[i*2*65+j+k+3]=(short)RightSample;
-                                        RightIMACode>>=4;
-                                    }
-                                    IMAData+=2;
-                                }
-                            }
-                            memset(&(ALBuf->data[(size/72*2*65)]), 0, 32);
-                            ALBuf->size=size/72*2*65*sizeof(ALshort);
-                            ALBuf->frequency=freq;
-                        }
-                        else
-                            alSetError(AL_OUT_OF_MEMORY);
-                    }
-                    else
                         alSetError(AL_INVALID_VALUE);
-                    break;
+                        break;
+                    }
+
+                    size /= 36;
+                    size *= 65;
+
+                    // Allocate extra padding samples
+                    temp = realloc(ALBuf->data, (padding*OrigChans + size)*sizeof(ALshort));
+                    if(temp)
+                    {
+                        ALBuf->data = temp;
+                        ConvertDataIMA4(ALBuf->data, data, OrigChans, size/65);
+
+                        memset(&(ALBuf->data[size]), 0, padding*sizeof(ALshort)*OrigChans);
+
+                        ALBuf->format = ((OrigChans==1) ? AL_FORMAT_MONO16 : AL_FORMAT_STEREO16);
+                        ALBuf->eOriginalFormat = format;
+                        ALBuf->size = size*sizeof(ALshort);
+                        ALBuf->frequency = freq;
+                        ALBuf->padding = padding;
+                    }
+                    else
+                        alSetError(AL_OUT_OF_MEMORY);
+                }   break;
 
                 default:
                     alSetError(AL_INVALID_ENUM);
@@ -600,6 +403,110 @@ ALAPI ALvoid ALAPIENTRY alBufferData(ALuint buffer,ALenum format,const ALvoid *d
     ProcessContext(Context);
 }
 
+/*
+*    alBufferSubDataEXT(ALuint buffer,ALenum format,ALvoid *data,ALsizei offset,ALsizei length)
+*
+*    Fill buffer with audio data
+*/
+ALvoid ALAPIENTRY alBufferSubDataEXT(ALuint buffer,ALenum format,const ALvoid *data,ALsizei offset,ALsizei length)
+{
+    ALCcontext *Context;
+    ALbuffer *ALBuf;
+
+    Context = alcGetCurrentContext();
+    SuspendContext(Context);
+
+    if(alIsBuffer(buffer) && buffer != 0)
+    {
+        ALBuf = (ALbuffer*)ALTHUNK_LOOKUPENTRY(buffer);
+        if(ALBuf->data == NULL)
+        {
+            // buffer does not have any data
+            alSetError(AL_INVALID_NAME);
+        }
+        else if(length < 0 || offset < 0 || (length > 0 && data == NULL))
+        {
+            // data is NULL or offset/length is negative
+            alSetError(AL_INVALID_VALUE);
+        }
+        else
+        {
+            switch(format)
+            {
+                case AL_FORMAT_REAR8:
+                case AL_FORMAT_REAR16:
+                case AL_FORMAT_REAR32: {
+                    ALuint OrigBytes = ((format==AL_FORMAT_REAR8) ? 1 :
+                                        ((format==AL_FORMAT_REAR16) ? 2 :
+                                         4));
+
+                    if(ALBuf->eOriginalFormat != AL_FORMAT_REAR8 &&
+                       ALBuf->eOriginalFormat != AL_FORMAT_REAR16 &&
+                       ALBuf->eOriginalFormat != AL_FORMAT_REAR32)
+                    {
+                        alSetError(AL_INVALID_ENUM);
+                        break;
+                    }
+
+                    if(ALBuf->size/4/sizeof(ALshort) < (ALuint)offset+length)
+                    {
+                        alSetError(AL_INVALID_VALUE);
+                        break;
+                    }
+
+                    ConvertDataRear(&ALBuf->data[offset*4], data, OrigBytes, length*2);
+                }   break;
+
+                case AL_FORMAT_MONO_IMA4:
+                case AL_FORMAT_STEREO_IMA4: {
+                    int Channels = aluChannelsFromFormat(ALBuf->format);
+
+                    if(ALBuf->eOriginalFormat != format)
+                    {
+                        alSetError(AL_INVALID_ENUM);
+                        break;
+                    }
+
+                    if((offset%65) != 0 || (length%65) != 0 ||
+                       ALBuf->size/Channels/sizeof(ALshort) < (ALuint)offset+length)
+                    {
+                        alSetError(AL_INVALID_VALUE);
+                        break;
+                    }
+
+                    ConvertDataIMA4(&ALBuf->data[offset*Channels], data, Channels, length/65*Channels);
+                }   break;
+
+                default: {
+                    ALuint Channels = aluChannelsFromFormat(format);
+                    ALuint Bytes = aluBytesFromFormat(format);
+
+                    if(Channels != aluChannelsFromFormat(ALBuf->format))
+                    {
+                        alSetError(AL_INVALID_ENUM);
+                        break;
+                    }
+
+                    if(ALBuf->size/Channels/sizeof(ALshort) < (ALuint)offset+length)
+                    {
+                        alSetError(AL_INVALID_VALUE);
+                        break;
+                    }
+
+                    ConvertData(&ALBuf->data[offset*Channels], data, Bytes, length*Channels);
+                }   break;
+            }
+        }
+    }
+    else
+    {
+        // Invalid Buffer Name
+        alSetError(AL_INVALID_NAME);
+    }
+
+    ProcessContext(Context);
+}
+
 
 ALAPI void ALAPIENTRY alBufferf(ALuint buffer, ALenum eParam, ALfloat flValue)
 {
@@ -997,7 +904,8 @@ static void LoadData(ALbuffer *ALBuf, const ALubyte *data, ALsizei size, ALuint
     ALuint NewChannels = aluChannelsFromFormat(NewFormat);
     ALuint OrigBytes = aluBytesFromFormat(OrigFormat);
     ALuint OrigChannels = aluChannelsFromFormat(OrigFormat);
-    ALsizei i;
+    ALsizei padding = 2;
+    ALvoid *temp;
 
     assert(aluBytesFromFormat(NewFormat) == 2);
     assert(NewChannels == OrigChannels);
@@ -1008,79 +916,153 @@ static void LoadData(ALbuffer *ALBuf, const ALubyte *data, ALsizei size, ALuint
         return;
     }
 
-    switch(OrigBytes)
+    // Samples are converted to 16 bit here
+    size /= OrigBytes;
+    temp = realloc(ALBuf->data, (padding*NewChannels + size) * sizeof(ALshort));
+    if(temp)
     {
-    case 1:
-        size /= sizeof(ALubyte);
+        ALBuf->data = temp;
+        ConvertData(ALBuf->data, data, OrigBytes, size);
 
-        // 8bit Samples are converted to 16 bit here
-        // Allocate 8 extra samples
-        ALBuf->data = realloc(ALBuf->data, (8*NewChannels + size) * (1*sizeof(ALshort)));
-        if (ALBuf->data)
-        {
-            for (i = 0;i < size;i++)
-                ALBuf->data[i] = (ALshort)((data[i]-128) << 8);
-            memset(&(ALBuf->data[size]), 0, 16*NewChannels);
+        memset(&(ALBuf->data[size]), 0, padding*NewChannels*sizeof(ALshort));
 
-            ALBuf->format = NewFormat;
-            ALBuf->eOriginalFormat = OrigFormat;
-            ALBuf->size = size*1*sizeof(ALshort);
-            ALBuf->frequency = freq;
-        }
-        else
-            alSetError(AL_OUT_OF_MEMORY);
-        break;
+        ALBuf->format = NewFormat;
+        ALBuf->eOriginalFormat = OrigFormat;
+        ALBuf->size = size*sizeof(ALshort);
+        ALBuf->frequency = freq;
+        ALBuf->padding = padding;
+    }
+    else
+        alSetError(AL_OUT_OF_MEMORY);
+}
 
-    case 2:
-        size /= sizeof(ALshort);
+static void ConvertData(ALshort *dst, const ALvoid *src, ALint origBytes, ALsizei len)
+{
+    ALsizei i;
+    switch(origBytes)
+    {
+        case 1:
+            for(i = 0;i < len;i++)
+                dst[i] = ((ALshort)((ALubyte*)src)[i] - 128) << 8;
+            break;
 
-        // Allocate 8 extra samples
-        ALBuf->data = realloc(ALBuf->data, (8*NewChannels + size) * (1*sizeof(ALshort)));
-        if (ALBuf->data)
-        {
-            memcpy(ALBuf->data, data, size*1*sizeof(ALshort));
-            memset(&(ALBuf->data[size]), 0, 16*NewChannels);
+        case 2:
+            memcpy(dst, src, len*sizeof(ALshort));
+            break;
 
-            ALBuf->format = NewFormat;
-            ALBuf->eOriginalFormat = OrigFormat;
-            ALBuf->size = size*1*sizeof(ALshort);
-            ALBuf->frequency = freq;
-        }
-        else
-            alSetError(AL_OUT_OF_MEMORY);
-        break;
-
-    case 4:
-        size /= sizeof(ALfloat);
-
-        // Allocate 8 extra samples
-        ALBuf->data = realloc(ALBuf->data, (8*NewChannels + size) * (1*sizeof(ALshort)));
-        if (ALBuf->data)
-        {
-            ALint smp;
-            for (i = 0;i < size;i++)
+        case 4:
+            for(i = 0;i < len;i++)
             {
-                smp = (((ALfloat*)data)[i] * 32767.5f - 0.5f);
+                ALint smp;
+                smp = (((ALfloat*)src)[i] * 32767.5f - 0.5f);
                 smp = min(smp,  32767);
                 smp = max(smp, -32768);
-                ALBuf->data[i] = (ALshort)smp;
+                dst[i] = (ALshort)smp;
             }
-            memset(&(ALBuf->data[size]), 0, 16*NewChannels);
+            break;
 
-            ALBuf->format = NewFormat;
-            ALBuf->eOriginalFormat = OrigFormat;
-            ALBuf->size = size*1*sizeof(ALshort);
-            ALBuf->frequency = freq;
-        }
-        else
-            alSetError(AL_OUT_OF_MEMORY);
-        break;
-
-    default:
-        assert(0);
+        default:
+            assert(0);
     }
 }
 
+static void ConvertDataRear(ALshort *dst, const ALvoid *src, ALint origBytes, ALsizei len)
+{
+    ALsizei i;
+    switch(origBytes)
+    {
+        case 1:
+            for(i = 0;i < len;i+=4)
+            {
+                dst[i+0] = 0;
+                dst[i+1] = 0;
+                dst[i+2] = ((ALshort)((ALubyte*)src)[i/2+0] - 128) << 8;
+                dst[i+3] = ((ALshort)((ALubyte*)src)[i/2+1] - 128) << 8;
+            }
+            break;
+
+        case 2:
+            for(i = 0;i < len;i+=4)
+            {
+                dst[i+0] = 0;
+                dst[i+1] = 0;
+                dst[i+2] = ((ALshort*)src)[i/2+0];
+                dst[i+3] = ((ALshort*)src)[i/2+1];
+            }
+            break;
+
+        case 4:
+            for(i = 0;i < len;i+=4)
+            {
+                ALint smp;
+                dst[i+0] = 0;
+                dst[i+1] = 0;
+                smp = (((ALfloat*)src)[i/2+0] * 32767.5f - 0.5);
+                smp = min(smp,  32767);
+                smp = max(smp, -32768);
+                dst[i+2] = (ALshort)smp;
+                smp = (((ALfloat*)src)[i/2+1] * 32767.5f - 0.5);
+                smp = min(smp,  32767);
+                smp = max(smp, -32768);
+                dst[i+3] = (ALshort)smp;
+            }
+            break;
+
+        default:
+            assert(0);
+    }
+}
+
+static void ConvertDataIMA4(ALshort *dst, const ALvoid *src, ALint origChans, ALsizei len)
+{
+    const ALuint *IMAData;
+    ALint Sample[2],Index[2];
+    ALuint IMACode[2];
+    ALsizei i,j,k,c;
+
+    assert(origChans <= 2);
+
+    IMAData = src;
+    for(i = 0;i < len/origChans;i++)
+    {
+        for(c = 0;c < origChans;c++)
+        {
+            Sample[c] = ((ALshort*)IMAData)[0];
+            Index[c] = ((ALshort*)IMAData)[1];
+
+            Index[c] = ((Index[c]<0) ? 0 : Index[c]);
+            Index[c] = ((Index[c]>88) ? 88 : Index[c]);
+
+            dst[i*65*origChans + c] = (ALshort)Sample[c];
+
+            IMAData++;
+        }
+
+        for(j = 1;j < 65;j += 8)
+        {
+            for(c = 0;c < origChans;c++)
+                IMACode[c] = *(IMAData++);
+
+            for(k = 0;k < 8;k++)
+            {
+                for(c = 0;c < origChans;c++)
+                {
+                    Sample[c] += ((g_IMAStep_size[Index[c]]*g_IMACodeword_4[IMACode[c]&15])/8);
+                    Index[c] += g_IMAIndex_adjust_4[IMACode[c]&15];
+
+                    if(Sample[c] < -32768) Sample[c] = -32768;
+                    else if(Sample[c] > 32767) Sample[c] = 32767;
+
+                    if(Index[c]<0) Index[c] = 0;
+                    else if(Index[c]>88) Index[c] = 88;
+
+                    dst[(i*65+j+k)*origChans + c] = (ALshort)Sample[c];
+                    IMACode[c] >>= 4;
+                }
+            }
+        }
+    }
+}
 
 /*
 *    ReleaseALBuffers()
@@ -1094,7 +1076,7 @@ ALvoid ReleaseALBuffers(ALvoid)
 
 #ifdef _DEBUG
     if(g_uiBufferCount > 0)
-        AL_PRINT("exit() %d Buffer(s) NOT deleted\n", g_uiBufferCount);
+        AL_PRINT("exit(): deleting %d Buffer(s)\n", g_uiBufferCount);
 #endif
 
     ALBuffer = g_pBuffers;
