@@ -30,6 +30,7 @@
 #include "alEffect.h"
 #include "alAuxEffectSlot.h"
 #include "alSource.h"
+#include "alBuffer.h"
 #include "AL/al.h"
 #include "AL/alc.h"
 
@@ -144,6 +145,8 @@ static ALfunction  function[]=   {
     { "alGetAuxiliaryEffectSlotf",  (ALvoid *) alGetAuxiliaryEffectSlotf },
     { "alGetAuxiliaryEffectSlotfv", (ALvoid *) alGetAuxiliaryEffectSlotfv},
 
+    { "alBufferSubDataEXT",         (ALvoid *) alBufferSubDataEXT        },
+
     { NULL,                         (ALvoid *) NULL                      } };
 
 static ALenums enumeration[]={
