@@ -32,8 +32,8 @@
 #include "alThunk.h"
 #include "alAuxEffectSlot.h"
 
-static ALvoid InitSourceParams(ALsource *pSource);
-static ALboolean GetSourceOffset(ALsource *pSource, ALenum eName, ALfloat *pflOffset);
+static ALvoid InitSourceParams(ALCcontext *Context, ALsource *pSource);
+static ALboolean GetSourceOffset(ALsource *pSource, ALenum eName, ALfloat *pflOffset, ALuint updateSize);
 static ALvoid ApplyOffset(ALsource *pSource, ALboolean bUpdateContext);
 static ALint GetByteOffset(ALsource *pSource);
 
@@ -78,7 +78,7 @@ ALAPI ALvoid ALAPIENTRY alGenSources(ALsizei n,ALuint *sources)
                             sources[i] = (ALuint)ALTHUNK_ADDENTRY(*list);
                             (*list)->source = sources[i];
 
-                            InitSourceParams(*list);
+                            InitSourceParams(Context, *list);
                             Context->SourceCount++;
                             i++;
 
@@ -137,73 +137,64 @@ ALAPI ALvoid ALAPIENTRY alDeleteSources(ALsizei n, const ALuint *sources)
 
             if (Device)
             {
-                if ((ALuint)n <= Context->SourceCount)
+                // Check that all Sources are valid (and can therefore be deleted)
+                for (i = 0; i < n; i++)
                 {
-                    // Check that all Sources are valid (and can therefore be deleted)
-                    for (i = 0; i < n; i++)
+                    if (!alIsSource(sources[i]))
                     {
-                        if (!alIsSource(sources[i]))
-                        {
-                            alSetError(AL_INVALID_NAME);
-                            bSourcesValid = AL_FALSE;
-                            break;
-                        }
-                    }
-
-                    if (bSourcesValid)
-                    {
-                        // All Sources are valid, and can be deleted
-                        for (i = 0; i < n; i++)
-                        {
-                            // Recheck that the Source is valid, because there could be duplicated Source names
-                            if (alIsSource(sources[i]))
-                            {
-                                ALSource=((ALsource *)ALTHUNK_LOOKUPENTRY(sources[i]));
-                                alSourceStop((ALuint)ALSource->source);
-
-                                // For each buffer in the source's queue, decrement its reference counter and remove it
-                                while (ALSource->queue != NULL)
-                                {
-                                    ALBufferList = ALSource->queue;
-                                    // Decrement buffer's reference counter
-                                    if (ALBufferList->buffer != 0)
-                                        ((ALbuffer*)(ALTHUNK_LOOKUPENTRY(ALBufferList->buffer)))->refcount--;
-                                    // Update queue to point to next element in list
-                                    ALSource->queue = ALBufferList->next;
-                                    // Release memory allocated for buffer list item
-                                    free(ALBufferList);
-                                }
-
-                                for(j = 0;j < MAX_SENDS;++j)
-                                {
-                                    if(ALSource->Send[j].Slot)
-                                        ALSource->Send[j].Slot->refcount--;
-                                    ALSource->Send[j].Slot = NULL;
-                                }
-
-                                // Decrement Source count
-                                Context->SourceCount--;
-
-                                // Remove Source from list of Sources
-                                list = &Context->Source;
-                                while(*list && *list != ALSource)
-                                    list = &(*list)->next;
-
-                                if(*list)
-                                    *list = (*list)->next;
-                                ALTHUNK_REMOVEENTRY(ALSource->source);
-
-                                memset(ALSource,0,sizeof(ALsource));
-                                free(ALSource);
-                            }
-                        }
-
+                        alSetError(AL_INVALID_NAME);
+                        bSourcesValid = AL_FALSE;
+                        break;
                     }
                 }
-                else
+
+                if (bSourcesValid)
                 {
-                    // Trying to delete more Sources than have been generated
-                    alSetError(AL_INVALID_NAME);
+                    // All Sources are valid, and can be deleted
+                    for (i = 0; i < n; i++)
+                    {
+                        // Recheck that the Source is valid, because there could be duplicated Source names
+                        if (alIsSource(sources[i]))
+                        {
+                            ALSource=((ALsource *)ALTHUNK_LOOKUPENTRY(sources[i]));
+                            alSourceStop((ALuint)ALSource->source);
+
+                            // For each buffer in the source's queue, decrement its reference counter and remove it
+                            while (ALSource->queue != NULL)
+                            {
+                                ALBufferList = ALSource->queue;
+                                // Decrement buffer's reference counter
+                                if (ALBufferList->buffer != 0)
+                                    ((ALbuffer*)(ALTHUNK_LOOKUPENTRY(ALBufferList->buffer)))->refcount--;
+                                // Update queue to point to next element in list
+                                ALSource->queue = ALBufferList->next;
+                                // Release memory allocated for buffer list item
+                                free(ALBufferList);
+                            }
+
+                            for(j = 0;j < MAX_SENDS;++j)
+                            {
+                                if(ALSource->Send[j].Slot)
+                                    ALSource->Send[j].Slot->refcount--;
+                                ALSource->Send[j].Slot = NULL;
+                            }
+
+                            // Decrement Source count
+                            Context->SourceCount--;
+
+                            // Remove Source from list of Sources
+                            list = &Context->Source;
+                            while(*list && *list != ALSource)
+                                list = &(*list)->next;
+
+                            if(*list)
+                                *list = (*list)->next;
+                            ALTHUNK_REMOVEENTRY(ALSource->source);
+
+                            memset(ALSource,0,sizeof(ALsource));
+                            free(ALSource);
+                        }
+                    }
                 }
             }
             else
@@ -368,12 +359,19 @@ ALAPI ALvoid ALAPIENTRY alSourcef(ALuint source, ALenum eParam, ALfloat flValue)
                 break;
 
             case AL_ROOM_ROLLOFF_FACTOR:
-                if (flValue >= 0.0f && flValue <= 1.0f)
+                if (flValue >= 0.0f && flValue <= 10.0f)
                     pSource->RoomRolloffFactor = flValue;
                 else
                     alSetError(AL_INVALID_VALUE);
                 break;
 
+            case AL_DOPPLER_FACTOR:
+                if (flValue >= 0.0f && flValue <= 1.0f)
+                    pSource->DopplerFactor = flValue;
+                else
+                    alSetError(AL_INVALID_VALUE);
+                break;
+
             case AL_SEC_OFFSET:
             case AL_SAMPLE_OFFSET:
             case AL_BYTE_OFFSET:
@@ -634,9 +632,6 @@ ALAPI ALvoid ALAPIENTRY alSourcei(ALuint source,ALenum eParam,ALint lValue)
                             pSource->lSourceType = AL_UNDETERMINED;
                         }
 
-                        // Set Buffers Processed
-                        pSource->BuffersProcessed = 0;
-
                         // Update AL_BUFFER parameter
                         pSource->ulBufferID = lValue;
                     }
@@ -709,6 +704,19 @@ ALAPI ALvoid ALAPIENTRY alSourcei(ALuint source,ALenum eParam,ALint lValue)
                     alSetError(AL_INVALID_VALUE);
                 break;
 
+            case AL_DISTANCE_MODEL:
+                if(lValue == AL_NONE ||
+                   lValue == AL_INVERSE_DISTANCE ||
+                   lValue == AL_INVERSE_DISTANCE_CLAMPED ||
+                   lValue == AL_LINEAR_DISTANCE ||
+                   lValue == AL_LINEAR_DISTANCE_CLAMPED ||
+                   lValue == AL_EXPONENT_DISTANCE ||
+                   lValue == AL_EXPONENT_DISTANCE_CLAMPED)
+                    pSource->DistanceModel = lValue;
+                else
+                    alSetError(AL_INVALID_VALUE);
+                break;
+
             default:
                 alSetError(AL_INVALID_ENUM);
                 break;
@@ -824,6 +832,7 @@ ALAPI void ALAPIENTRY alSourceiv(ALuint source, ALenum eParam, const ALint* plVa
                 case AL_DIRECT_FILTER_GAINHF_AUTO:
                 case AL_AUXILIARY_SEND_FILTER_GAIN_AUTO:
                 case AL_AUXILIARY_SEND_FILTER_GAINHF_AUTO:
+                case AL_DISTANCE_MODEL:
                     alSourcei(source, eParam, plValues[0]);
                     break;
 
@@ -858,7 +867,7 @@ ALAPI ALvoid ALAPIENTRY alGetSourcef(ALuint source, ALenum eParam, ALfloat *pflV
 {
     ALCcontext    *pContext;
     ALsource    *pSource;
-    ALfloat        flOffset;
+    ALfloat        flOffset[2];
 
     pContext = alcGetCurrentContext();
     if (pContext)
@@ -908,8 +917,20 @@ ALAPI ALvoid ALAPIENTRY alGetSourcef(ALuint source, ALenum eParam, ALfloat *pflV
                 case AL_SEC_OFFSET:
                 case AL_SAMPLE_OFFSET:
                 case AL_BYTE_OFFSET:
-                    if (GetSourceOffset(pSource, eParam, &flOffset))
-                        *pflValue = flOffset;
+                    if(GetSourceOffset(pSource, eParam, flOffset, pContext->Device->UpdateSize))
+                        *pflValue = flOffset[0];
+                    else
+                        alSetError(AL_INVALID_OPERATION);
+                    break;
+
+                case AL_SEC_RW_OFFSETS_EXT:
+                case AL_SAMPLE_RW_OFFSETS_EXT:
+                case AL_BYTE_RW_OFFSETS_EXT:
+                    if(GetSourceOffset(pSource, eParam, flOffset, pContext->Device->UpdateSize))
+                    {
+                        pflValue[0] = flOffset[0];
+                        pflValue[1] = flOffset[1];
+                    }
                     else
                         alSetError(AL_INVALID_OPERATION);
                     break;
@@ -934,6 +955,10 @@ ALAPI ALvoid ALAPIENTRY alGetSourcef(ALuint source, ALenum eParam, ALfloat *pflV
                     *pflValue = pSource->RoomRolloffFactor;
                     break;
 
+                case AL_DOPPLER_FACTOR:
+                    *pflValue = pSource->DopplerFactor;
+                    break;
+
                 default:
                     alSetError(AL_INVALID_ENUM);
                     break;
@@ -1034,6 +1059,7 @@ ALAPI ALvoid ALAPIENTRY alGetSourcefv(ALuint source, ALenum eParam, ALfloat *pfl
                 case AL_MAX_GAIN:
                 case AL_MAX_DISTANCE:
                 case AL_ROLLOFF_FACTOR:
+                case AL_DOPPLER_FACTOR:
                 case AL_CONE_OUTER_GAIN:
                 case AL_SEC_OFFSET:
                 case AL_SAMPLE_OFFSET:
@@ -1089,7 +1115,7 @@ ALAPI ALvoid ALAPIENTRY alGetSourcei(ALuint source, ALenum eParam, ALint *plValu
 {
     ALCcontext *pContext;
     ALsource   *pSource;
-    ALfloat     flOffset;
+    ALfloat     flOffset[2];
 
     pContext = alcGetCurrentContext();
     if (pContext)
@@ -1152,7 +1178,7 @@ ALAPI ALvoid ALAPIENTRY alGetSourcei(ALuint source, ALenum eParam, ALint *plValu
                         *plValue = 0;
                     }
                     else
-                        *plValue = pSource->BuffersProcessed;
+                        *plValue = pSource->BuffersPlayed;
                     break;
 
                 case AL_SOURCE_TYPE:
@@ -1162,8 +1188,20 @@ ALAPI ALvoid ALAPIENTRY alGetSourcei(ALuint source, ALenum eParam, ALint *plValu
                 case AL_SEC_OFFSET:
                 case AL_SAMPLE_OFFSET:
                 case AL_BYTE_OFFSET:
-                    if (GetSourceOffset(pSource, eParam, &flOffset))
-                        *plValue = (ALint)flOffset;
+                    if(GetSourceOffset(pSource, eParam, flOffset, pContext->Device->UpdateSize))
+                        *plValue = (ALint)flOffset[0];
+                    else
+                        alSetError(AL_INVALID_OPERATION);
+                    break;
+
+                case AL_SEC_RW_OFFSETS_EXT:
+                case AL_SAMPLE_RW_OFFSETS_EXT:
+                case AL_BYTE_RW_OFFSETS_EXT:
+                    if(GetSourceOffset(pSource, eParam, flOffset, pContext->Device->UpdateSize))
+                    {
+                        plValue[0] = (ALint)flOffset[0];
+                        plValue[1] = (ALint)flOffset[1];
+                    }
                     else
                         alSetError(AL_INVALID_OPERATION);
                     break;
@@ -1184,6 +1222,14 @@ ALAPI ALvoid ALAPIENTRY alGetSourcei(ALuint source, ALenum eParam, ALint *plValu
                     *plValue = pSource->WetGainHFAuto;
                     break;
 
+                case AL_DOPPLER_FACTOR:
+                    *plValue = (ALint)pSource->DopplerFactor;
+                    break;
+
+                case AL_DISTANCE_MODEL:
+                    *plValue = pSource->DistanceModel;
+                    break;
+
                 default:
                     alSetError(AL_INVALID_ENUM);
                     break;
@@ -1291,12 +1337,14 @@ ALAPI void ALAPIENTRY alGetSourceiv(ALuint source, ALenum eParam, ALint* plValue
                 case AL_BYTE_OFFSET:
                 case AL_MAX_DISTANCE:
                 case AL_ROLLOFF_FACTOR:
+                case AL_DOPPLER_FACTOR:
                 case AL_REFERENCE_DISTANCE:
                 case AL_SOURCE_TYPE:
                 case AL_DIRECT_FILTER:
                 case AL_DIRECT_FILTER_GAINHF_AUTO:
                 case AL_AUXILIARY_SEND_FILTER_GAIN_AUTO:
                 case AL_AUXILIARY_SEND_FILTER_GAINHF_AUTO:
+                case AL_DISTANCE_MODEL:
                     alGetSourcei(source, eParam, plValues);
                     break;
 
@@ -1351,7 +1399,7 @@ ALAPI ALvoid ALAPIENTRY alSourcePlayv(ALsizei n, const ALuint *pSourceList)
     ALbufferlistitem    *ALBufferList;
     ALboolean            bSourcesValid = AL_TRUE;
     ALboolean            bPlay;
-    ALsizei              i;
+    ALsizei              i, j;
 
     pContext = alcGetCurrentContext();
     if (pContext)
@@ -1394,6 +1442,11 @@ ALAPI ALvoid ALAPIENTRY alSourcePlayv(ALsizei n, const ALuint *pSourceList)
 
                     if (bPlay)
                     {
+                        for(j = 0;j < OUTPUTCHANNELS;j++)
+                            pSource->DryGains[j] = 0.0f;
+                        for(j = 0;j < MAX_SENDS;j++)
+                            pSource->WetGains[j] = 0.0f;
+
                         if (pSource->state != AL_PAUSED)
                         {
                             pSource->state = AL_PLAYING;
@@ -1401,10 +1454,8 @@ ALAPI ALvoid ALAPIENTRY alSourcePlayv(ALsizei n, const ALuint *pSourceList)
                             pSource->play = AL_TRUE;
                             pSource->position = 0;
                             pSource->position_fraction = 0;
-                            pSource->BuffersProcessed = 0;
                             pSource->BuffersPlayed = 0;
-                            pSource->BufferPosition = 0;
-                            pSource->lBytesPlayed = 0;
+                            pSource->FirstStart = AL_TRUE;
 
                             pSource->ulBufferID = pSource->queue->buffer;
 
@@ -1421,6 +1472,7 @@ ALAPI ALvoid ALAPIENTRY alSourcePlayv(ALsizei n, const ALuint *pSourceList)
                             pSource->state = AL_PLAYING;
                             pSource->inuse = AL_TRUE;
                             pSource->play = AL_TRUE;
+                            pSource->FirstStart = AL_FALSE;
                         }
 
                         // Check if an Offset has been set
@@ -1437,7 +1489,7 @@ ALAPI ALvoid ALAPIENTRY alSourcePlayv(ALsizei n, const ALuint *pSourceList)
                             ALBufferList = ALBufferList->next;
                         }
 
-                        pSource->BuffersPlayed = pSource->BuffersProcessed = pSource->BuffersInQueue;
+                        pSource->BuffersPlayed = pSource->BuffersInQueue;
                     }
                 }
             }
@@ -1561,7 +1613,7 @@ ALAPI ALvoid ALAPIENTRY alSourceStopv(ALsizei n, const ALuint *sources)
                     {
                         Source->state=AL_STOPPED;
                         Source->inuse=AL_FALSE;
-                        Source->BuffersPlayed = Source->BuffersProcessed = Source->BuffersInQueue;
+                        Source->BuffersPlayed = Source->BuffersInQueue;
                         ALBufferListItem= Source->queue;
                         while (ALBufferListItem != NULL)
                         {
@@ -1633,7 +1685,7 @@ ALAPI ALvoid ALAPIENTRY alSourceRewindv(ALsizei n, const ALuint *sources)
                         Source->inuse=AL_FALSE;
                         Source->position=0;
                         Source->position_fraction=0;
-                        Source->BuffersProcessed = 0;
+                        Source->BuffersPlayed = 0;
                         ALBufferListItem= Source->queue;
                         while (ALBufferListItem != NULL)
                         {
@@ -1906,7 +1958,6 @@ ALAPI ALvoid ALAPIENTRY alSourceUnqueueBuffers( ALuint source, ALsizei n, ALuint
                     // Release memory for buffer list item
                     free(ALBufferList);
                     ALSource->BuffersInQueue--;
-                    ALSource->BuffersProcessed--;
                 }
 
                 if (ALSource->state != AL_PLAYING)
@@ -1920,10 +1971,7 @@ ALAPI ALvoid ALAPIENTRY alSourceUnqueueBuffers( ALuint source, ALsizei n, ALuint
                 }
 
                 if((ALuint)n > ALSource->BuffersPlayed)
-                {
                     ALSource->BuffersPlayed = 0;
-                    ALSource->BufferPosition = 0;
-                }
                 else
                     ALSource->BuffersPlayed -= n;
             }
@@ -1951,7 +1999,7 @@ ALAPI ALvoid ALAPIENTRY alSourceUnqueueBuffers( ALuint source, ALsizei n, ALuint
 }
 
 
-static ALvoid InitSourceParams(ALsource *pSource)
+static ALvoid InitSourceParams(ALCcontext *Context, ALsource *pSource)
 {
     pSource->flInnerAngle = 360.0f;
     pSource->flOuterAngle = 360.0f;
@@ -1980,6 +2028,9 @@ static ALvoid InitSourceParams(ALsource *pSource)
     pSource->WetGainHFAuto = AL_TRUE;
     pSource->AirAbsorptionFactor = 0.0f;
     pSource->RoomRolloffFactor = 0.0f;
+    pSource->DopplerFactor = 1.0f;
+
+    pSource->DistanceModel = Context->DistanceModel;
 
     pSource->state = AL_INITIAL;
     pSource->lSourceType = AL_UNDETERMINED;
@@ -1994,12 +2045,13 @@ static ALvoid InitSourceParams(ALsource *pSource)
     Gets the current playback position in the given Source, in the appropriate format (Bytes, Samples or MilliSeconds)
     The offset is relative to the start of the queue (not the start of the current buffer)
 */
-static ALboolean GetSourceOffset(ALsource *pSource, ALenum eName, ALfloat *pflOffset)
+static ALboolean GetSourceOffset(ALsource *pSource, ALenum eName, ALfloat *pflOffset, ALuint updateSize)
 {
     ALbufferlistitem *pBufferList;
     ALbuffer         *pBuffer;
     ALfloat        flBufferFreq;
-    ALint        lBytesPlayed, lChannels;
+    ALint        lChannels;
+    ALint        readPos, writePos;
     ALenum        eOriginalFormat;
     ALboolean    bReturn = AL_TRUE;
     ALint        lTotalBufferDataSize;
@@ -2013,15 +2065,20 @@ static ALboolean GetSourceOffset(ALsource *pSource, ALenum eName, ALfloat *pflOf
         lChannels = aluChannelsFromFormat(pBuffer->format);
 
         // Get Current BytesPlayed
-        lBytesPlayed = pSource->position * lChannels * 2; // NOTE : This is the byte offset into the *current* buffer
+        readPos = pSource->position * lChannels * 2; // NOTE : This is the byte offset into the *current* buffer
         // Add byte length of any processed buffers in the queue
         pBufferList = pSource->queue;
         while ((pBufferList) && (pBufferList->bufferstate == PROCESSED))
         {
-            lBytesPlayed += ((ALbuffer*)ALTHUNK_LOOKUPENTRY(pBufferList->buffer))->size;
+            readPos += ((ALbuffer*)ALTHUNK_LOOKUPENTRY(pBufferList->buffer))->size;
             pBufferList = pBufferList->next;
         }
 
+        if(pSource->state == AL_PLAYING)
+            writePos = readPos + (updateSize * lChannels * 2);
+        else
+            writePos = readPos;
+
         lTotalBufferDataSize = 0;
         pBufferList = pSource->queue;
         while (pBufferList)
@@ -2033,64 +2090,88 @@ static ALboolean GetSourceOffset(ALsource *pSource, ALenum eName, ALfloat *pflOf
 
         if (pSource->bLooping)
         {
-            if (lBytesPlayed < 0)
-                lBytesPlayed = 0;
+            if(readPos < 0)
+                readPos = 0;
             else
-                lBytesPlayed = lBytesPlayed % lTotalBufferDataSize;
+                readPos %= lTotalBufferDataSize;
+            if(writePos < 0)
+                writePos = 0;
+            else
+                writePos %= lTotalBufferDataSize;
         }
         else
         {
             // Clamp BytesPlayed to within 0 and lTotalBufferDataSize
-            if(lBytesPlayed < 0)
-                lBytesPlayed = 0;
-            if(lBytesPlayed > lTotalBufferDataSize)
-                lBytesPlayed = lTotalBufferDataSize;
+            if(readPos < 0)
+                readPos = 0;
+            else if(readPos > lTotalBufferDataSize)
+                readPos = lTotalBufferDataSize;
+            if(writePos < 0)
+                writePos = 0;
+            else if(writePos > lTotalBufferDataSize)
+                writePos = lTotalBufferDataSize;
         }
 
         switch (eName)
         {
         case AL_SEC_OFFSET:
-            *pflOffset = ((ALfloat)lBytesPlayed / (lChannels * 2.0f * flBufferFreq));
+        case AL_SEC_RW_OFFSETS_EXT:
+            pflOffset[0] = (ALfloat)readPos / (lChannels * 2.0f * flBufferFreq);
+            pflOffset[1] = (ALfloat)writePos / (lChannels * 2.0f * flBufferFreq);
             break;
         case AL_SAMPLE_OFFSET:
-            *pflOffset = (ALfloat)(lBytesPlayed / (lChannels * 2));
+        case AL_SAMPLE_RW_OFFSETS_EXT:
+            pflOffset[0] = (ALfloat)(readPos / (lChannels * 2));
+            pflOffset[1] = (ALfloat)(writePos / (lChannels * 2));
             break;
         case AL_BYTE_OFFSET:
+        case AL_BYTE_RW_OFFSETS_EXT:
             // Take into account the original format of the Buffer
             if ((eOriginalFormat == AL_FORMAT_MONO_IMA4) ||
                 (eOriginalFormat == AL_FORMAT_STEREO_IMA4))
             {
-                // Compression rate of the ADPCM supported is 3.6111 to 1
-                lBytesPlayed = (ALint)((ALfloat)lBytesPlayed / 3.6111f);
                 // Round down to nearest ADPCM block
-                *pflOffset = (ALfloat)((lBytesPlayed / (36 * lChannels)) * 36 * lChannels);
+                pflOffset[0] = (ALfloat)((readPos / (65 * 2 * lChannels)) * 36 * lChannels);
+                if(pSource->state == AL_PLAYING)
+                {
+                    // Round up to nearest ADPCM block
+                    pflOffset[1] = (ALfloat)(((writePos + (65 * 2 * lChannels) - 1) / (65 * 2 * lChannels)) * 36 * lChannels);
+                }
+                else
+                    pflOffset[1] = pflOffset[0];
             }
             else if (eOriginalFormat == AL_FORMAT_REAR8)
             {
-                *pflOffset = (ALfloat)(lBytesPlayed >> 2);
+                pflOffset[0] = (ALfloat)(readPos >> 2);
+                pflOffset[1] = (ALfloat)(writePos >> 2);
             }
             else if (eOriginalFormat == AL_FORMAT_REAR16)
             {
-                *pflOffset = (ALfloat)(lBytesPlayed >> 1);
+                pflOffset[0] = (ALfloat)(readPos >> 1);
+                pflOffset[1] = (ALfloat)(writePos >> 1);
             }
             else if (aluBytesFromFormat(eOriginalFormat) == 1)
             {
-                *pflOffset = (ALfloat)(lBytesPlayed >> 1);
+                pflOffset[0] = (ALfloat)(readPos >> 1);
+                pflOffset[1] = (ALfloat)(writePos >> 1);
             }
             else if (aluBytesFromFormat(eOriginalFormat) == 4)
             {
-                *pflOffset = (ALfloat)(lBytesPlayed << 1);
+                pflOffset[0] = (ALfloat)(readPos << 1);
+                pflOffset[1] = (ALfloat)(writePos << 1);
             }
             else
             {
-                *pflOffset = (ALfloat)lBytesPlayed;
+                pflOffset[0] = (ALfloat)readPos;
+                pflOffset[1] = (ALfloat)writePos;
             }
             break;
         }
     }
     else
     {
-        *pflOffset = 0.0f;
+        pflOffset[0] = 0.0f;
+        pflOffset[1] = 0.0f;
     }
 
     return bReturn;
@@ -2120,7 +2201,6 @@ static void ApplyOffset(ALsource *pSource, ALboolean bUpdateContext)
         pBufferList = pSource->queue;
         lTotalBufferSize = 0;
         pSource->BuffersPlayed = 0;
-        pSource->BuffersProcessed = 0;
         while (pBufferList)
         {
             pBuffer = ALTHUNK_LOOKUPENTRY(pBufferList->buffer);
@@ -2133,10 +2213,7 @@ static void ApplyOffset(ALsource *pSource, ALboolean bUpdateContext)
                 pSource->BuffersPlayed++;
 
                 if (!pSource->bLooping)
-                {
                     pBufferList->bufferstate = PROCESSED;
-                    pSource->BuffersProcessed++;
-                }
             }
             else if (lTotalBufferSize <= lByteOffset)
             {
@@ -2146,14 +2223,8 @@ static void ApplyOffset(ALsource *pSource, ALboolean bUpdateContext)
                 // Set Current Buffer ID
                 pSource->ulBufferID = pBufferList->buffer;
 
-                // Set current position in this buffer
-                pSource->BufferPosition = lByteOffset - lTotalBufferSize;
-
-                // Set Total Bytes Played to Offset
-                pSource->lBytesPlayed = lByteOffset;
-
                 // SW Mixer Positions are in Samples
-                pSource->position = pSource->BufferPosition /
+                pSource->position = (lByteOffset - lTotalBufferSize) /
                                     aluBytesFromFormat(pBuffer->format) /
                                     aluChannelsFromFormat(pBuffer->format);
             }
@@ -2288,7 +2359,7 @@ ALvoid ReleaseALSources(ALCcontext *Context)
 {
 #ifdef _DEBUG
     if(Context->SourceCount > 0)
-        AL_PRINT("alcDestroyContext(): %d Source(s) NOT deleted\n", Context->SourceCount);
+        AL_PRINT("alcDestroyContext(): deleting %d Source(s)\n", Context->SourceCount);
 #endif
 
     while(Context->Source)
